@@ -1,18 +1,24 @@
+import configparser
+import copy
 import os.path
 from configparser import ConfigParser
 from io import StringIO
+from textwrap import dedent
 
 import pytest
 
 from configupdater import (
+    AlreadyAttachedError,
     Comment,
     ConfigUpdater,
     DuplicateOptionError,
     DuplicateSectionError,
     MissingSectionHeaderError,
     NoConfigFileReadError,
+    NoneValueDisallowed,
     NoOptionError,
     NoSectionError,
+    NotAttachedError,
     Option,
     ParsingError,
     Section,
@@ -58,6 +64,25 @@ def test_update_no_cfg():
         updater.update_file()
 
 
+def test_update_file_issue_68(tmp_path):
+    # allow_no_value = False
+    file = tmp_path / "file.cfg"
+    file.write_text("[section]")
+    cfg = ConfigUpdater(allow_no_value=False).read_file(open(file))
+    cfg.set("section", "option")
+    with pytest.warns(NoneValueDisallowed):
+        # A warning is issued, but the method will not fail
+        cfg.update_file()
+    assert file.read_text().strip() == "[section]"
+
+    # allow_no_value = True
+    file.write_text("[section]")
+    cfg = ConfigUpdater(allow_no_value=True).read_file(open(file))
+    cfg.set("section", "option")
+    cfg.update_file()
+    assert file.read_text().strip() == "[section]\noption"
+
+
 def test_str(setup_cfg_path, setup_cfg):
     updater = ConfigUpdater()
     updater.read(setup_cfg_path)
@@ -108,8 +133,11 @@ def test_len_updater(setup_cfg_path):
 def test_iter_section(setup_cfg_path):
     updater = ConfigUpdater()
     updater.read(setup_cfg_path)
-    # we test the number of blocks, not sections
-    assert len([block for block in updater]) == 14
+    # iter_blocks will give us everything, iter_sections just the section objects and
+    # __iter__ should work as iter_blocks
+    assert len([block for block in updater.iter_blocks()]) == 14
+    assert len([block for block in updater.iter_sections()]) == 12
+    assert len([entry for entry in updater]) == 12
 
 
 def test_iter_items_section(setup_cfg_path):
@@ -137,8 +165,11 @@ def test_get_section(setup_cfg_path):
     assert section._structure
     with pytest.raises(KeyError):
         updater["non_existent_section"]
-    with pytest.raises(ValueError):
-        updater._get_section_idx("non_existent_section")
+
+    # `get_section` is equivalent to dict.get
+    assert updater.get_section("non_existent_section") is None
+    default = {}
+    assert updater.get_section("non_existent_section", default) is default
 
 
 def test_section_to_dict(setup_cfg_path):
@@ -186,8 +217,11 @@ def test_iter_option(setup_cfg_path):
     updater = ConfigUpdater()
     updater.read(setup_cfg_path)
     section = updater["metadata"]
-    # we test the number of entries, not options
-    assert len([entry for entry in section]) == 12
+    # iter_blocks will give us everything, iter_options just the option objects and
+    # __iter__ should work as iter_blocks
+    assert len([entry for entry in section.iter_blocks()]) == 12
+    assert len([entry for entry in section.iter_options()]) == 9
+    assert len([entry for entry in section]) == 9
 
 
 def test_get_options(setup_cfg_path):
@@ -199,6 +233,12 @@ def test_get_options(setup_cfg_path):
     with pytest.raises(NoSectionError):
         updater.options("non_existent_section")
 
+    # Section objects also have a "get" method similar to dict
+    section = updater["metadata"]
+    assert section.get("name").value == "configupdater"
+    assert section.get("non_existent_option") is None
+    assert section.get("non_existent_option", []) == []
+
 
 def test_items(setup_cfg_path):
     updater = ConfigUpdater()
@@ -275,6 +315,7 @@ key =
 def test_get_option():
     updater = ConfigUpdater()
     updater.read_string(test1_cfg_in)
+    updater.read_string(test1_cfg_in)
     option = updater["default"]["key"]
     assert option.value == "1"
     with pytest.raises(KeyError):
@@ -307,7 +348,7 @@ def test_remove_option():
 
 
 def test_set_option():
-    updater = ConfigUpdater()
+    updater = ConfigUpdater(allow_no_value=True)
     updater.read_string(test1_cfg_in)
     updater.set("default", "key", "1")
     assert updater["default"]["key"].value == "1"
@@ -327,10 +368,17 @@ def test_set_option():
     assert values == ["param1", "param2"]  # non destructive operation
     with pytest.raises(NoSectionError):
         updater.set("wrong_section", "key", "1")
+    new_option = copy.deepcopy(updater["default"]["key"])
+    updater["default"]["key"] = new_option
+    assert updater["default"]["key"] is new_option
+    new_option = copy.deepcopy(updater["default"]["key"])
+    new_option.key = "wrong_key"
+    with pytest.raises(ValueError):
+        updater["default"]["key"] = new_option
 
 
 def test_section_set_option():
-    updater = ConfigUpdater()
+    updater = ConfigUpdater(allow_no_value=True)
     updater.read_string(test1_cfg_in)
     default_sec = updater["default"]
     default_sec.set("key", "1")
@@ -384,8 +432,6 @@ def test_del_section():
     assert str(updater) == test2_cfg_out
     with pytest.raises(KeyError):
         del updater["section2"]
-    with pytest.raises(ValueError):
-        updater["section1"]._get_option_idx("wrong key")
 
 
 test_wrong_cfg = """
@@ -401,12 +447,21 @@ def test_handle_error():
 
 
 def test_validate_format(setup_cfg_path):
-    updater = ConfigUpdater(allow_no_value=False)
+    updater = ConfigUpdater(allow_no_value=True)
     updater.read(setup_cfg_path)
-    updater.validate_format()
+    updater.validate_format(allow_no_value=False)
     updater.set("metadata", "author")
     with pytest.raises(ParsingError):
-        updater.validate_format()
+        updater.validate_format(allow_no_value=False)
+
+
+def test_validate_format_propagates_kwargs(setup_cfg_path):
+    updater = ConfigUpdater(allow_no_value=True)
+    updater.read(setup_cfg_path)
+    updater.set("pyscaffold", "namespace")
+    assert updater["pyscaffold"]["namespace"].value is None
+    assert str(updater["pyscaffold"]["namespace"]) == "namespace\n"  # no `=` sign
+    assert updater.validate_format() is True
 
 
 test3_cfg_in = """
@@ -444,7 +499,7 @@ key2
 
 
 def test_add_before_after_option():
-    updater = ConfigUpdater()
+    updater = ConfigUpdater(allow_no_value=True)
     updater.read_string(test4_cfg_in)
     with pytest.raises(ValueError):
         updater["section"].add_before.option("key0", 0)
@@ -531,7 +586,7 @@ def test_add_before_after_section():
     sect_updater.read_string(test6_cfg_in)
     section = sect_updater["section0"]
     section.name = "new_section"
-    updater["section2"].add_after.section(section)
+    updater["section2"].add_after.section(section.detach())
     assert str(updater) == test6_cfg_out_new_sect
     with pytest.raises(DuplicateSectionError):
         updater["section2"].add_after.section(section)
@@ -586,14 +641,17 @@ def test_set_item_section():
         updater["section"] = "newsection"
     sect_updater.read_string(test6_cfg_in)
     section = sect_updater["section0"]
-    updater["new_section"] = section
+    with pytest.raises(ValueError):
+        updater["new_section"] = section
+    section.name = "new_section"
+    updater["new_section"] = section.detach()
     assert str(updater) == test6_cfg_out_new_sect
     # test overwriting an existing section
     updater.read_string(test6_cfg_in)
     sect_updater.read_string(test6_cfg_in)
     exist_section = sect_updater["section0"]
     exist_section["key0"] = 42
-    updater["section0"] = exist_section
+    updater["section0"] = exist_section.detach()
     assert str(updater) == test6_cfg_out_overwritten
 
 
@@ -940,6 +998,11 @@ def test_empty_lines_in_values_support():
     assert updater["main1"]["key1"].value == parser["main1"]["key1"]
     assert updater["main2"]["key1"].value == parser["main2"]["key1"]
     assert updater["main4"]["key1"].value == parser["main4"]["key2"]
+
+    # Make sure values are not being stored in "whitespace nodes"
+    value = (x.strip() for x in updater["main1"]["key1"].value.splitlines())
+    assert [v for v in value if v] == ["a", "b", "c", "d"]
+
     assert test21_cfg_in == str(updater)
     with pytest.raises(ParsingError):
         updater = ConfigUpdater(empty_lines_in_values=False)
@@ -974,6 +1037,299 @@ def test_navigation_and_remove():
     assert isinstance(key2.next_block, Space)
     assert key2.next_block is section.last_block
     assert section.last_block.next_block is None
-    key1.next_block.remove()
-    key2.next_block.remove()
+    key1.next_block.detach()
+    key2.next_block.detach()
     assert str(updater) == test22_cfg_out
+
+
+test23_cfg_in = """
+[section1]
+    # indented comment
+a = 42
+"""
+
+
+def test_indented_comment():
+    updater = ConfigUpdater()
+    updater.read_string(test23_cfg_in)
+    section = updater["section1"]
+    assert isinstance(section.first_block, Comment)
+    assert str(updater) == test23_cfg_in
+
+
+test24_cfg_in = """
+[sec1]
+key1 = val1
+key2 = val2
+
+[sec2]
+key1 = val1
+key2 = val2
+"""
+
+
+def test_iter_consistency_with_configparser():
+    parser = ConfigParser()
+    parser.read_string(test24_cfg_in)
+    updater = ConfigUpdater()
+    updater.read_string(test24_cfg_in)
+    # [1:] to drop 'DEFAULT' section
+    assert list(parser)[1:] == list(updater)
+    assert list(parser["sec1"]) == list(updater["sec1"])
+
+
+def test_add_detached_section_option_objects():
+    updater = ConfigUpdater()
+    updater.read_string(test24_cfg_in)
+    sec1 = updater["sec1"]
+    sec2 = updater["sec2"]
+    assert sec2.container is updater
+    sec2.detach()
+    assert not updater.has_section("sec2")
+    assert not sec2.has_container()
+    with pytest.raises(NotAttachedError):
+        sec2.container
+
+    new_sec2 = Section("new-sec2")
+    new_opt = Option(key="new-key", value="new-value")
+    new_sec2.add_option(new_opt)
+    with pytest.raises(AlreadyAttachedError):
+        sec1.add_option(new_opt)
+    updater.add_section(new_sec2)
+    assert updater.has_section("new-sec2")
+    assert updater["new-sec2"]["new-key"].value == "new-value"
+
+    new_sec3 = Section("new-sec3")
+    new_opt2 = Option(key="new-key", value="new-value")
+    updater["new-sec3"] = new_sec3
+    new_sec3["new-key"] = new_opt2
+
+
+def test_transferring_blocks_between_elements():
+    # Let's say a user have a big new section that it wants to insert in an existing
+    # document. Instead of programmatically creating this new section using the builder
+    # API, they might be tempted to parse this new section from template files.
+
+    existing = """\
+    [section0]
+    option0 = 0
+    """
+
+    template1 = """\
+    [section1]
+    option1 = 1
+    """
+
+    template2 = """\
+    [section2]
+    option2 = 2
+    # comment
+    """
+
+    target = ConfigUpdater()
+    target.read_string(dedent(existing))
+
+    source1 = ConfigUpdater()
+    source1.read_string(dedent(template1))
+
+    source2 = ConfigUpdater()
+    source2.read_string(dedent(template2))
+
+    target["section1"] = source1["section1"].detach()
+    assert "section1" in target
+
+    target["section1"].add_after.section(source2["section2"].detach())
+    assert "section2" not in source1
+    assert "section2" in target
+
+
+def test_transfering_sections_and_manipulating_options():
+    # This test case is similar to `test_transfering_sections_between_elements`,
+    # but now we use the builder API in the options inside the transferred section.
+    # We need to make sure that when a section is copied to a different tree,
+    # the references inside the options are updated, so we don't end up adding blocks in
+    # the original object.
+
+    existing = """\
+    [section0]
+    option0 = 0
+    """
+
+    template1 = """\
+    [section1]
+    option1 = 1
+    """
+
+    target = ConfigUpdater()
+    target.read_string(dedent(existing))
+
+    source1 = ConfigUpdater()
+    source1.read_string(dedent(template1))
+
+    target["section1"] = source1["section1"].detach()
+    assert "section1" in target
+
+    target["section1"]["option1"].add_after.option("option2", "value")
+
+    assert "option2" not in str(source1)
+    assert "option2" in str(target)
+
+
+def test_transfering_sections_and_manipulating_options_with_deepcopy():
+    existing = """\
+    [section0]
+    option0 = 0
+    """
+
+    template1 = """\
+    [section1]
+    option1 = 1
+    """
+
+    template2 = """\
+    [section2]
+    option2 = 2
+    # comment
+    """
+
+    target = ConfigUpdater()
+    target.read_string(dedent(existing))
+
+    source1 = ConfigUpdater()
+    source1.read_string(dedent(template1))
+
+    source2 = ConfigUpdater()
+    source2.read_string(dedent(template2))
+
+    target["section1"] = copy.deepcopy(source1["section1"])
+    assert "section1" in target
+    assert "section1" in source1
+
+    target["section1"].add_after.section(copy.deepcopy(source2["section2"]))
+    target["section2"]["option2"] = "value"
+    assert "option2 = value" in str(target)
+    assert "option2" not in str(source1)
+    assert "value" not in str(source2)
+
+
+def test_deepcopy():
+    example = """\
+    [options]
+    testing =   # Add here test requirements (used by tox)
+        sphinx  # required for system tests
+        flake8  # required for system tests
+    """
+    updater = ConfigUpdater().read_string(dedent(example))
+    clone = copy.deepcopy(updater)
+    assert clone["options"].container is clone
+    assert clone["options"]["testing"].container is not updater["options"]
+    clone["options"]["testing"].detach()
+    assert "testing" not in str(clone)
+    assert "testing" in str(updater)
+    clone.add_section("other")
+    assert "other" in clone
+    assert "other" not in updater
+    assert str(updater) == dedent(example)
+
+
+def test_section_comment():
+    section = """\
+    [section0] # section comment
+    key = value
+    """
+
+    updater = ConfigUpdater()
+    updater.read_string(dedent(section))
+
+    updater["section0"].name = "section1"
+
+    expected = """\
+    [section1] # section comment
+    key = value
+    """
+    assert str(updater) == dedent(expected)
+    assert updater["section1"].raw_comment == " # section comment"
+
+    updater["section1"].raw_comment = " # new comment"
+    expected = """\
+    [section1] # new comment
+    key = value
+    """
+    assert str(updater) == dedent(expected)
+
+
+def test_setitem_detached_option():
+    existing = """\
+    [section0]
+    option0 = 0
+    option1 = # No value
+    """
+
+    template1 = """\
+    [section1]
+    option1 = 1
+    """
+
+    target = ConfigUpdater()
+    target.read_string(dedent(existing))
+
+    source1 = ConfigUpdater()
+    source1.read_string(dedent(template1))
+
+    option1 = source1["section1"]["option1"].detach()
+    target["section0"]["option1"] = option1
+    assert target["section0"]["option1"].value == "1"
+
+
+def test_comments_in_multiline_values():
+    indented_case = """\
+    [header]
+    install_requires =
+        importlib-metadata; python_version<"3.8"
+        #Adding some comments here that are perfectly valid.
+        some-other-dependency
+    """
+    unindented_case = """\
+    [header]
+    install_requires =
+        importlib-metadata; python_version<"3.8"
+    #Adding some comments here that are perfectly valid.
+        some-other-dependency
+    """
+    error_case_if_no_value = """\
+    [header]
+    install_requires =
+        importlib-metadata; python_version<"3.8"
+    #Adding some comments here that are perfectly valid.
+    some-other-dependency
+    """
+    parser = ConfigParser(allow_no_value=False)
+    updater = ConfigUpdater(allow_no_value=False)
+
+    parser.read_string(dedent(indented_case))
+    updater.read_string(dedent(indented_case))
+    assert str(updater) == dedent(indented_case)
+
+    parser.read_string(dedent(unindented_case))
+    updater.read_string(dedent(unindented_case))
+    assert str(updater) == dedent(unindented_case)
+
+    with pytest.raises(configparser.ParsingError):
+        parser.read_string(dedent(error_case_if_no_value))
+    with pytest.raises(ParsingError):
+        updater.read_string(dedent(error_case_if_no_value))
+
+    parser = ConfigParser(allow_no_value=True)
+    updater = ConfigUpdater(allow_no_value=True)
+
+    parser.read_string(dedent(indented_case))
+    updater.read_string(dedent(indented_case))
+    assert str(updater) == dedent(indented_case)
+
+    parser.read_string(dedent(unindented_case))
+    updater.read_string(dedent(unindented_case))
+    assert str(updater) == dedent(unindented_case)
+
+    parser.read_string(dedent(error_case_if_no_value))
+    updater.read_string(dedent(error_case_if_no_value))
+    assert str(updater) == dedent(error_case_if_no_value)
